(* A graph is a list of nodes that are connected together. We can
   represent a graph using the following data structures. *)

type name  = string

type node = {
  label:name;              (* Name of the node *)
  value:int;               (* Value that it stores *)
  mutable edges:node list  (* List of nodes that neighbor it. *)
}

type graph = node list 

(* A sample graph with 6 nodes:

   a -> d -> e 
   ^    |
   |    |
   V    V 
   b -> c  -> f

   For example, we can represent the graph above with the following code.
   Since graphs may have cycles in them, we must make this list of edges
   mutable.
*)

let test_graph : graph = 
  let a = { label="a"; value=1; edges = [] } in
  let b = { label="b"; value=2; edges = [] } in
  let c = { label="c"; value=3; edges = [] } in
  let d = { label="d"; value=4; edges = [] } in
  let e = { label="e"; value=5; edges = [] } in
  let f = { label="f"; value=6; edges = [] } in
    a.edges <- [b;d];
    b.edges <- [a;c];
    c.edges <- [f];
    d.edges <- [c;e];
    [a;b;c;d;e;f]

(* ----------------- Depth first search of the graph --------------------- *)
type path = name list (* A path is a list of labels describing the route. *)

type cont = Halt | EdgeCont of name * cont 
            | AuxCont of node list * name * cont

let rec visit (l:name) (c:cont) : bool = 
  match c with 
      Halt -> false 
    | EdgeCont (n,old_cont) -> l=n || visit l old_cont
    | AuxCont (_,_,old_cont) -> visit l old_cont


(* depth-first search. Find a path from the starting node
   to a node labelled with stop. *)
let rec dfs (start:node) (stop:name) : path option = 
  dfs_aux start stop Halt

and dfs_aux 
  (current:node) (* the node that we're currently starting the path from *)
  (stop:name)    (* the label of the node that we are looking for. *)
  (cont:cont) 
  : path option = 
  let l = current.label in
    print_string (l ^ " ");  (* Print out the order of traversal of nodes *)
  if l = stop then apply_cont cont (Some [l])
  else if visit l cont then apply_cont cont (None)
  else dfs_edges current.edges stop (EdgeCont (l, cont))
    
and dfs_edges (edges:node list)(stop:name) (cont:cont): path option = 
  match edges with 
      [] -> apply_cont cont None
    | (hd::tl) -> 
        dfs_aux hd stop (AuxCont (tl,stop,cont))

and apply_cont (cont:cont)(p:path option) : path option = 
  match cont with 
      Halt -> p
    | EdgeCont (l, old_cont) -> 
        (match p with 
            Some p -> apply_cont old_cont (Some (l::p))
          | None -> apply_cont old_cont None)
    | AuxCont (tl,stop,old_cont) -> 
        (match p with 
             Some p -> apply_cont old_cont (Some p)
           | None -> dfs_edges tl stop old_cont)
      
(* --------- code to print out the answer once we've found it ----------- *)

let rec string_of_path (p:path) : string = 
  match p with 
      [] -> ""
    | [l] -> l
    | (hd::tl) -> hd ^ " -> " ^ (string_of_path tl)
  
let doit (start:name) : unit = 
  print_string "Searching: ";
  let po = dfs (List.hd test_graph) start in 
  print_string "\nPATH IS: ";
  match po with 
      Some p -> (print_string (string_of_path p); print_string "\n")
    | None -> print_string "NO PATH\n"
        
let main = 
  doit "c";
  doit "e"
