(* A graph is a list of nodes that are connected together. We can
   represent a graph using the following data structures. *)

type name  = string

type node = {
  label:name;              (* Name of the node *)
  value:int;               (* Value that it stores *)
  mutable edges:node list  (* List of nodes that neighbor it. *)
}

type graph = node list 

(* A sample graph with 6 nodes:

   a -> d -> e 
   ^    |
   |    |
   V    V 
   b -> c  -> f

   For example, we can represent the graph above with the following code.
   Since graphs may have cycles in them, we must make this list of edges
   mutable.
*)

let test_graph : graph = 
  let a = { label="a"; value=1; edges = [] } in
  let b = { label="b"; value=2; edges = [] } in
  let c = { label="c"; value=3; edges = [] } in
  let d = { label="d"; value=4; edges = [] } in
  let e = { label="e"; value=5; edges = [] } in
  let f = { label="f"; value=6; edges = [] } in
    a.edges <- [b;d];
    b.edges <- [a;c];
    c.edges <- [f];
    d.edges <- [c;e];
    [a;b;c;d;e;f]

(* ----------------- Depth first search of the graph --------------------- *)
type path = name list (* A path is a list of labels describing the route. *)

type cont = Halt | Aux of node list * name * name list * cont

(* depth-first search. Find a path from the starting node
   to a node labelled with stop. *)
let rec dfs (start:node) (stop:name) : int option = 
  dfs_aux start stop [] Halt

and dfs_aux 
  (current:node) (* the node that we're currently starting the path from *)
  (stop:name)           (* the label of the node that we are looking for. *)
  (visited:name list)   (* a list of nodes that we have seen on this path, 
                           so that we may detect cycles *)
  (cont:cont)
  : int option = 
  let l = current.label in
    print_string (l ^ " ");  (* Print out the order of traversal of nodes *)
  if l = stop then apply_cont Halt (Some current.value)
  else if List.mem l visited then apply_cont cont None
  else dfs_edges current.edges stop (l::visited) cont 

and dfs_edges (edges:node list)(stop:name)(visited:name list) (cont:cont): int option = 
  match edges with 
      [] -> apply_cont cont None
    | (hd::tl) -> dfs_aux hd stop visited (Aux(tl,stop, visited,cont))
            
and apply_cont (cont:cont) (v:int option) : int option = 
  match cont with 
      Halt -> v
    | Aux (tl,stop,visited,cont) ->  dfs_edges tl stop visited cont
        (* Don't have to look at v because we
           know that it will be "None" *)


(* --------- code to print out the answer once we've found it ----------- *)

let doit (start:name) : unit = 
  print_string "Searching: ";
  let po = dfs (List.hd test_graph) start in 
  print_string "\nVALUE IS: ";
  match po with 
      Some v -> (print_string (string_of_int v); print_string "\n")
    | None -> print_string "NO VALUE\n"
        
let main = 
  doit "c";
  doit "e";
  doit "g"

